using System;

namespace Littlechip.Tools
{
	/// <summary>
	/// Provides symmetric encryption/decryption services.
	/// </summary>
	public sealed class Cryptography
	{
		#region Encrypt
		/// <summary>
		/// Encrypts the given text using the system default algorithm, key, and initialization vector.
		/// This really should not be used unless the default key is more secure than the key that you might pass in.
		/// </summary>
		/// <param name="text">Text to encrypt.</param>
		/// <returns>Encrypted text.</returns>
		public static string Encrypt(string text)
		{
			return Encrypt(text, GetStandardKey(), GetStandardIV(), GetStandardAlgorithm());
		}

		/// <summary>
		/// Encrypts the given text using the given key.  Uses system default algorithm and initialization vector.
		/// </summary>
		/// <param name="text">Text to encrypt.</param>
		/// <param name="key">Key to use in encryption.</param>
		/// <returns>Encrypted text.</returns>
		public static string Encrypt(string text, string key)
		{
			return Encrypt(text, key, GetStandardIV(), GetStandardAlgorithm());
		}

		/// <summary>
		/// Encrypts the given text using the given key and initialization vector.  Uses system default algorithm.
		/// </summary>
		/// <param name="text">Text to encrypt.</param>
		/// <param name="key">Key to use in encryption.</param>
		/// <param name="IV">Initialization vector for the algorithm.</param>
		/// <returns>Encrypted text.</returns>
		public static string Encrypt(string text, string key, string IV)
		{
			return Encrypt(text, key, IV, GetStandardAlgorithm());
		}

		/// <summary>
		/// Encrypts the given text using the given key, initialization vector, and algorithm.
		/// </summary>
		/// <param name="text">Text to encrypt.</param>
		/// <param name="key">Key to use in encryption.</param>
		/// <param name="IV">Initialization vector for the algorithm.</param>
		/// <param name="algorithmName">Name of the algorithm to use.</param>
		/// <returns>Encrypted text.</returns>
		public static string Encrypt(string text, string key, string IV, string algorithmName)
		{
			byte[] keyBytes;
			byte[] ivBytes;
			byte[] inTextBytes;
			byte[] outTextBytes;
			System.Text.StringBuilder ret;

			inTextBytes = System.Text.Encoding.ASCII.GetBytes(text);
			ivBytes = System.Text.Encoding.ASCII.GetBytes(IV);
			keyBytes = System.Text.Encoding.ASCII.GetBytes(GetLegalKey(key, algorithmName));
			System.Security.Cryptography.SymmetricAlgorithm csp = System.Security.Cryptography.SymmetricAlgorithm.Create(algorithmName);
			System.Security.Cryptography.ICryptoTransform xForm = csp.CreateEncryptor(keyBytes, ivBytes);
			outTextBytes = xForm.TransformFinalBlock(inTextBytes, 0, inTextBytes.Length);
			
			// Get bytes as hex string
			ret = new System.Text.StringBuilder();
			foreach(byte b in outTextBytes)
			{
				//Format as hex
				ret.AppendFormat("{0:X2}", b);
			} 

			return ret.ToString(); // old way: System.Text.Encoding.ASCII.GetString(outTextBytes);
		}
		#endregion

		#region Decrypt
		/// <summary>
		/// Decrypts the given text using the system default algorithm, key, and initialization vector.
		/// </summary>
		/// <param name="text">Text to decrypt.</param>
		/// <returns>Decrypted text.</returns>
		public static string Decrypt(string text)
		{
			return Decrypt(text, GetStandardKey(), GetStandardIV(), GetStandardAlgorithm());
		}
		
		/// <summary>
		/// Decrypts the given text using the given key.  Uses system default algorithm and initialization vector.
		/// </summary>
		/// <param name="text">Text to decrypt.</param>
		/// <param name="key">Key to use in decryption.</param>
		/// <returns>Decrypted text.</returns>
		public static string Decrypt(string text, string key)
		{
			return Decrypt(text, key, GetStandardIV(), GetStandardAlgorithm());
		}

		/// <summary>
		/// Decrypts the given text using the given key and initialization vector.  Uses system default algorithm.
		/// </summary>
		/// <param name="text">Text to decrypt.</param>
		/// <param name="key">Key to use in decryption.</param>
		/// <param name="IV">Initialization vector for the algorithm.</param>
		/// <returns>Decrypted text.</returns>
		public static string Decrypt(string text, string key, string IV)
		{
			return Decrypt(text, key, IV, GetStandardAlgorithm());
		}

		/// <summary>
		/// Decrypts the given text using the given key, initialization vector, and algorithm.
		/// </summary>
		/// <param name="text">Text to decrypt.</param>
		/// <param name="key">Key to use in decryption.</param>
		/// <param name="IV">Initialization vector for the algorithm.</param>
		/// <param name="algorithmName">Name of the algorithm to use.</param>
		/// <returns>Decrypted text.</returns>
		public static string Decrypt(string text, string key, string IV, string algorithmName)
		{
			byte[] keyBytes;
			byte[] ivBytes;
			byte[] inTextBytes;
			byte[] outTextBytes;

			try
			{
				//Put the input string into the byte array
				inTextBytes = new byte[text.Length / 2];
				for(int x = 0; x < text.Length / 2; x++)
				{
					int i = (Convert.ToInt32(text.Substring(x * 2, 2), 16));
					inTextBytes[x] = (byte)i;
				}
			}
			catch (System.FormatException ex)
			{
				throw new FormatException("Text to decrypt has been corrupted or was not encrypted with this tool.", ex);
			}

			// old way: inTextBytes = System.Text.Encoding.ASCII.GetBytes(text);
			ivBytes = System.Text.Encoding.ASCII.GetBytes(IV);
			keyBytes = System.Text.Encoding.ASCII.GetBytes(GetLegalKey(key, algorithmName));
			System.Security.Cryptography.SymmetricAlgorithm csp = System.Security.Cryptography.SymmetricAlgorithm.Create(algorithmName);
			System.Security.Cryptography.ICryptoTransform xForm = csp.CreateDecryptor(keyBytes, ivBytes);
			outTextBytes = xForm.TransformFinalBlock(inTextBytes, 0, inTextBytes.Length);
			return System.Text.Encoding.ASCII.GetString(outTextBytes);
		}
		#endregion

		#region Helper Functions
		/// <summary>
		/// Gets the legal key sizes in byte length for the given algorithm.
		/// </summary>
		/// <param name="algorithmName">Algorithm to get key sizes for.</param>
		/// <returns>Legal key sizes in byte length for the given algorithm.</returns>
		private static int[] GetLegalKeySizes(string algorithmName)
		{
			System.Security.Cryptography.SymmetricAlgorithm csp = System.Security.Cryptography.SymmetricAlgorithm.Create(algorithmName);
			System.Security.Cryptography.KeySizes[] sizes = csp.LegalKeySizes;
			System.Collections.ArrayList sizeList = new System.Collections.ArrayList();
			System.Security.Cryptography.KeySizes ks;
			int curSize;
			int[] returnValue;

			// for each key sizes object, append valid key sizes to our array list
			for (int i = 0; i < sizes.Length; i++)
			{
				ks = sizes[i];
				curSize = ks.MinSize;
				// starting with the minimum size, append sizes by the valid interval until we reach the max size
				while (curSize <= ks.MaxSize)
				{
					sizeList.Add((curSize/8));
					curSize += ks.SkipSize;
				}
			}
			// init return value to proper length
			returnValue = new int[sizeList.Count];
			// add valid sizes to our return value
			for (int i = 0; i < sizeList.Count; i++)
				returnValue[i] = Convert.ToInt32(sizeList[i]);
			return returnValue;
		}


		/// <summary>
		/// Gets the system standard initialization vector.
		/// </summary>
		/// <returns>The system standard initialization vector.</returns>
		private static string GetStandardIV()
		{
			// arbitrary 16 byte IV
			// only use this with algorithms that expect an IV of 16 bytes, 
			// such as Rijndael, aka AES.
			return "JJ"; 
		}

		/// <summary>
		/// Gets the system standard key.  This really should not be used unless it is more secure than the key that you might pass in.
		/// </summary>
		/// <returns>The system standard key.</returns>
		private static string GetStandardKey()
		{
			return "C29983F7A4F48BABF87F797A#KE(A_DI"; // arbitrary key
		}

		/// <summary>
		/// Gets the system standard algorithm.
		/// </summary>
		/// <returns>The system standard algorithm.</returns>
		private static string GetStandardAlgorithm()
		{
			return "Rijndael"; // aka AES
		}


		/// <summary>
		/// Gets a legal sized key for the given algorithm.
		/// </summary>
		/// <param name="key">Key passed to check and, if needed, modify for valid length.</param>
		/// <param name="algorithmName">Name of algorithm to get key for.</param>
		/// <returns>A legal sized key for the given algorithm.</returns>
		private static string GetLegalKey(string key, string algorithmName)
		{
			string systemKey = "693FD4AD09C29983F7A4F48BABF87F797AB25FDFEE5B9B807FF2AF9F8103584EA7845220DCD313B82256EBCD594F56D28F1FF2DC87417378876F4DFB8E1B628D";
			int[] legalSizes = GetLegalKeySizes(algorithmName);
			int maxSize = 0;
			int keySize = key.Length;

			// check to see if it's a legal size and get max size
			for (int i = 0; i < legalSizes.Length; i++)
			{
				// if it is a legal size, return the key
				if (legalSizes[i] == keySize)
					return key;
				// check the current size against the current max and upgrade max if current is bigger
				if (legalSizes[i] > maxSize)
					maxSize = legalSizes[i];
			}

			// if the key is less than the max/desired size, append the right amount of the system key to the key to get the right size
			if (keySize < maxSize)
				return key + systemKey.Substring(keySize, (maxSize - keySize));
			else // trim off the excess bytes from the key
				return key.Substring(0,maxSize);
		}

		#endregion
	}
}
